package env;

import jason.asSyntax.*;
import jason.environment.TimeSteppedEnvironment;

import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.logging.Logger;
import java.util.Calendar;
import java.util.concurrent.atomic.AtomicInteger;

import tools.Beta;
import tools.Random;
import jason.asSyntax.directives.FunctionRegister;

public class PeerReviewEnv extends TimeSteppedEnvironment {   
	static {
		FunctionRegister.addFunction(Beta.class);
		FunctionRegister.addFunction(Random.class);
	}

	static Logger logger = Logger.getLogger(PeerReviewEnv.class.getName());

	private int nConferences;			/// Number of conferences
	private int nScientists;			/// Number of scientists
	private int maxValue;               /// Higher limit of the range for the value of the papers (e.g. [1, 10])
	private int maxValueRev;            /// Higher limit of the range for the review of the papers (e.g. [1, 5])
	private int nMonths = 12;			/// Number of periods a year is divided into
	private int month = 0;              /// Current month
	private int year;                   /// Current year
	private int n_years;                /// Amount of years to be simulated
	private int endYear;                /// Last year of the simulation
	private String logDirName;          /// Logging directory name
	private long seed;                  /// Seed of the random number generator
	
	private int typeOfStepIdx = 1;     /// Type of step: 0=c(onferences), 1=s(cientists)
	private Atom typeOfStep[] = {ASSyntax.createAtom("c"), ASSyntax.createAtom("s")};
	
	private AtomicInteger startsReceived = new AtomicInteger(0);           /// Agents ready for the MAS initialization
	private AtomicInteger endsReceived = new AtomicInteger(0);           /// Agents ready for the MAS finalization
	private boolean start = false;              /// Start flag activated when all agents are ready
	private AtomicInteger conferencesCelebrated = new AtomicInteger(0);  /// Conferences already celebrated during the current year

	private PrintWriter fconferences;	/// Conferences log file
	private PrintWriter fscientists;	/// Scientists log file
	
    @Override
	public void init(String[] args) {
		nConferences = Integer.parseInt(args[0]);
		nScientists = Integer.parseInt(args[1]);
		maxValue = Integer.parseInt(args[2]);	
		maxValueRev = Integer.parseInt(args[3]);	
		nMonths = Integer.parseInt(args[4]);	
		year = Integer.parseInt(args[6])-1;
		n_years = Integer.parseInt(args[7]);
		endYear = year + n_years;
		logDirName = args[8];
		seed = Long.parseLong(args[9]);
		
		setOverActionsPolicy(OverActionsPolicy.queue);
		
		// Set general beliefs
		Literal max_value = ASSyntax.createLiteral("max_value",ASSyntax.createNumber(maxValue));
  	    addPercept(max_value);		
		Literal max_value_rev = ASSyntax.createLiteral("max_value_rev",ASSyntax.createNumber(maxValueRev));
  	    addPercept(max_value_rev);		
		Literal n_scientists = ASSyntax.createLiteral("n_scientists",ASSyntax.createNumber(nScientists));
  	    addPercept(n_scientists);		
		Literal periods_x_year = ASSyntax.createLiteral("periods_x_year",ASSyntax.createNumber(nMonths));
  	    addPercept(periods_x_year);		

		// Open log files
		File folder=new File("./logs");
		folder.mkdir(); 
		String dateFolder;
        if (logDirName.equals("null"))
		{
			Calendar date = Calendar.getInstance();
			dateFolder = date.get(Calendar.YEAR) + "-" + (date.get(Calendar.MONTH)+1) + "-" + date.get(Calendar.DAY_OF_MONTH) +
		          					 "_" + date.get(Calendar.HOUR_OF_DAY) + ":" + date.get(Calendar.MINUTE);   
		}
		else
		{
			dateFolder = logDirName;
		}
		folder=new File("./logs/" + dateFolder + "/");
		folder.mkdir();
		try
		{
			String header = "ConferenceName Month Year AccVa Succ PSub Pacc ReqQ AvgQ StdQ AvgRS StdRS FailA FailR";
			for (int i = 1; i <= maxValue; i++)
				if (i < 10)
					header += " PR-" + i;
				else
					header += " PR" + i;
			for (int i = 1; i <= maxValue; i++)
				if (i < 10)
					header += " PA-" + i;
				else
					header += " PA" + i;
			header += " 1/3Div 2/3Div AvgAQ StdAQ AvgAS StdAS AvgARS StdARS PCSz PxR N-Sc NRSc\n";

			fconferences = new PrintWriter(new FileWriter("./logs/" + dateFolder + "/conferences.txt"));
			fconferences.print(header);

			header = "ScientistName Type AthSk RevSk AvgQ Dis2Conf PxY NPapers NAcc NRej NPCs\n";
			fscientists = new PrintWriter(new FileWriter("./logs/" + dateFolder + "/scientists.txt"));
			fscientists.print(header);
		} catch (IOException e) {
			e.printStackTrace();
		}

		super.init(new String[] { args[5] } );
	}
    
	@Override
	public void stop()
	{
		super.stop();
		
		// Close log files
		fconferences.close();
		fscientists.close();
	}
	
	
	/**
	 * actions to execute at the beginning of an iteration before each agents step
	 *
	 * @param step	number of actual step
	 */
	@Override
	protected void stepStarted(int step) {
		if (step == 0)
		{
			// Initialize seeds of random number generators
			if (seed != 0)
			{
				Random.setSeed(seed);
				Beta.setSeed(seed);
			}
			
			addPercept(ASSyntax.createAtom("init"));
		}
		else 
			if ( start && (year <= endYear) )
			{
				// Delete previous date
				Literal l_date;
				if (typeOfStepIdx == 0)
				{
					l_date = ASSyntax.createLiteral("date", ASSyntax.createNumber(month), ASSyntax.createNumber(year), typeOfStep[typeOfStepIdx]);
				}
				else
				{
					if (month == 0)
					{
						l_date = ASSyntax.createLiteral("date", ASSyntax.createNumber(nMonths-1), ASSyntax.createNumber(year), typeOfStep[typeOfStepIdx]);
						year++;
					}
					else
						l_date = ASSyntax.createLiteral("date", ASSyntax.createNumber(month-1), ASSyntax.createNumber(year), typeOfStep[typeOfStepIdx]);
				}
				removePercept(l_date);
				//logger.severe("Remove date:" + l_date);
			
				// Set current date 
				typeOfStepIdx = (typeOfStepIdx+1) % 2;
				Literal l_new_date = ASSyntax.createLiteral("date", ASSyntax.createNumber(month), ASSyntax.createNumber(year), typeOfStep[typeOfStepIdx]);
				addPercept(l_new_date);		
				//logger.severe("New date:" + l_new_date);
	
				if (typeOfStepIdx == 1)
					month = (month+1)%nMonths;
			}
	}
	
	/**
	 * actions to execute before each agents step
	 *
	 * @param step		number of current step
	 * @param time		duration of the step
	 * @param timeout	if timeout has been exceeded
	 */
	@Override
    protected void stepFinished(int step, long time, boolean timeout) {
		if (!start && (startsReceived.get() == nConferences + nScientists) && (conferencesCelebrated.get() == 0) )
		{
			removePercept(ASSyntax.createAtom("init"));
			start = true;
			logger.severe("Start in step: " + step);
		}

        //if (timeout)
		//	logger.info("Timeout in step: " + step);
	}
	
    @Override
	public boolean executeAction(String ag, Structure action) 
	{
		//logger.info("process action: " + ag + " - " + action);
		
		if (action.getFunctor().equals("done") || 
		    action.getFunctor().equals("none")) {
			// Nothing to do
		}
		else if (action.getFunctor().equals("log_conference")){
			try {
				//logger.info("process action: " + ag + " - logconference");
				// Read parameters
				//log_conference(ConferenceName, CMonth, Year, AcceptValue, Success, NSubmitted, NAccepted, 
				//         		 math.average(Values), math.std_dev(Values), 
				//	     		 math.average(LSkills), math.std_dev(LSkills),
				//		 		 FailAccept, FailReject, OrdPaperDist, Divergence1, Divergence2, 
				//		 		 math.average(LAuthorAvgQ), math.std_dev(LAuthorAvgQ),
				//		 		 math.average(LAuthorAS), math.std_dev(LAuthorAS),
				//		 		 math.average(LAuthorRS), math.std_dev(LAuthorRS),
				//		 		 Csc, Crsc);
				String conferenceName = ((Atom)action.getTerm(0)).getFunctor();
				int log_month = (int)((NumberTerm)action.getTerm(1)).solve();
				int log_year = (int)((NumberTerm)action.getTerm(2)).solve();
				double accept_value = ((NumberTerm)action.getTerm(3)).solve();
				double success = ((NumberTerm)action.getTerm(4)).solve();
				int n_submitted_papers = (int)((NumberTerm)action.getTerm(5)).solve();
				int n_accepted_papers = (int)((NumberTerm)action.getTerm(6)).solve();
				double req_quality = ((NumberTerm)action.getTerm(7)).solve();
				double avg_values = ((NumberTerm)action.getTerm(8)).solve();
				double std_values = ((NumberTerm)action.getTerm(9)).solve();
				double avg_skills = ((NumberTerm)action.getTerm(10)).solve();
				double std_skills = ((NumberTerm)action.getTerm(11)).solve();
				int fail_accept = (int)((NumberTerm)action.getTerm(12)).solve();
				int fail_reject = (int)((NumberTerm)action.getTerm(13)).solve();
				double divergence1 = ((NumberTerm)action.getTerm(15)).solve();
				double divergence2 = ((NumberTerm)action.getTerm(16)).solve();
				double author_avg_q = ((NumberTerm)action.getTerm(17)).solve();
				double author_avg_q_std = ((NumberTerm)action.getTerm(18)).solve();
				double avg_author_skill = ((NumberTerm)action.getTerm(19)).solve();
				double std_author_skill = ((NumberTerm)action.getTerm(20)).solve();
				double avg_auth_rev_skill = ((NumberTerm)action.getTerm(21)).solve();
				double std_auth_rev_skill = ((NumberTerm)action.getTerm(22)).solve();
				int pc_size = (int)((NumberTerm)action.getTerm(23)).solve();
				int actual_paper_x_reviewer = (int)((NumberTerm)action.getTerm(24)).solve();
				int n_sc = (int)((NumberTerm)action.getTerm(25)).solve();
				int n_r_sc = (int)((NumberTerm)action.getTerm(26)).solve();
				
				// Save info
				fconferences.printf("%-14s %5d %4d %5.2f %4.2f %4d %4d %4.2f %4.2f %4.2f %5.2f %5.2f %5d %5d",
				                    conferenceName, log_month, log_year, accept_value, success, n_submitted_papers, n_accepted_papers,
				                    req_quality, avg_values, std_values, avg_skills, std_skills,
									fail_accept, fail_reject);
									
				for (Term t : (ListTermImpl) action.getTerm(14) ) 
					fconferences.printf(" %4d", (int)((NumberTerm)(((Literal)t).getTerm(1))).solve());
				for (Term t : (ListTermImpl) action.getTerm(14) ) 
					fconferences.printf(" %4d", (int)((NumberTerm)(((Literal)t).getTerm(2))).solve());
				fconferences.printf(" %6.2f %6.2f %5.2f %5.2f %5.2f %5.2f %6.2f %6.2f %4d %3d %4d %4d\n",
				                    divergence1, divergence2, author_avg_q, author_avg_q_std,
									avg_author_skill, std_author_skill, avg_auth_rev_skill, std_auth_rev_skill,
									pc_size, actual_paper_x_reviewer, n_sc, n_r_sc);
			 } catch (Exception e) {
				 e.printStackTrace();
				 return false;
			 }
				
			if (conferencesCelebrated.incrementAndGet() == nConferences*n_years)
			{ 
				start = false;
				Literal end = ASSyntax.createAtom("end");
				addPercept(end);		
			}

		} else if (action.getFunctor().equals("log_scientist")) {
			try {
				// Read parameters
				// log_scientist(MyName, Type, AuthorSkill, ReviewerSkill, AvgQ, MaxDist2Conf, 
				//               math.sum(LPapers), NA, NR, NPCs)
				String scientistName = ((Atom)action.getTerm(0)).getFunctor();
				String type = ((Atom)action.getTerm(1)).getFunctor();
				double authorSkill = ((NumberTerm)action.getTerm(2)).solve();
				double reviewerSkill = ((NumberTerm)action.getTerm(3)).solve();
				int avgQuality = (int)((NumberTerm)action.getTerm(4)).solve();
				int dist2conf = (int)((NumberTerm)action.getTerm(5)).solve();
				int papersPerYear = (int)((NumberTerm)action.getTerm(6)).solve();
				int nPapers = (int)((NumberTerm)action.getTerm(7)).solve();
				int nPapersAccepted = (int)((NumberTerm)action.getTerm(8)).solve();
				int nPapersRejected = (int)((NumberTerm)action.getTerm(9)).solve();
				int nPCMemberships = (int)((NumberTerm)action.getTerm(10)).solve();
			
				fscientists.printf("%-13s %-4s %5.2f %5.2f %4d %8d %3d %7d %4d %4d %4d\n",
				                  scientistName, type, authorSkill, reviewerSkill,
								  avgQuality, dist2conf, papersPerYear, nPapers, 
								  nPapersAccepted, nPapersRejected, nPCMemberships);
			} catch (Exception e) {
				e.printStackTrace();
				return false;
			}
		} 
		else if (action.getFunctor().equals("start")) {
			startsReceived.incrementAndGet();
		}
		else if (action.getFunctor().equals("end")) {
			if (endsReceived.incrementAndGet() == nConferences + nScientists)
			{
				try 
				{
					getEnvironmentInfraTier().getRuntimeServices().stopMAS();
				}
				catch (Exception e){
					e.printStackTrace();
					System.exit(1);
				}
			}
		}
		else {
		   	return false;
	    }

		return true;
	}
}
