package tools;

import jason.JasonException;
import jason.asSemantics.DefaultInternalAction;
import jason.asSemantics.InternalAction;
import jason.asSemantics.TransitionSystem;
import jason.asSemantics.Unifier;
import jason.asSyntax.NumberTerm;
import jason.asSyntax.Term;
import jason.asSyntax.ASSyntax;

import java.util.Iterator;
import java.util.Random;

/**
  <p>Internal action: <b><code>.random2(<i>Action</i>, <i>N</i>)</code></b>.
  
  <p>Description: wraps java.util.Random class and allows to generate different pseudorandom numbers.</p>
  
  <p>Parameter:<ul>
  
  <li>- action (atom): name of the fuction to be invoked (i.e. <code>setSeed</code>, <code>nextDouble</code>, <code>nextInt</code>, <code>nextBoolean</code> or <code>nextGaussian</code>)<br/>
  <li>- value (number): the variable to receive the random value (if action is <code>setSeed</code>, the new seed for the generator)<br/>
  <li><i>+ quantity of random numbers</i> (number, optional): default value is 1, value = 0 means that an infinify number of random numbers will be produced (this is useful for some backtrack circumstances).</li>
  
  </ul>
  
  <p>Example:<ul> 

  <li><code>.random2(setSeed, 1)</code>: sets 1 as the seed of the pseudorandom number generator.</li>
  <li><code>.random2(nextDouble, X)</code>: unifies X with one random number between 0 and 1.</li>
  <li><code>.random2(nextInt, X)</code>: unifies X with one random integer number, uniformly distributed  amongst all 2^32 possible int values.</li>
  <li><code>.random2(nextBoolean, X)</code>: unifies X with one random, uniformly distributed, boolean value (i.e. <code>true</code> or <code>false</code>).</li>
  <li><code>.random2(nextGaussian, X)</code>: unifies X with one Gaussian ("normally") distributed double value with mean 0.0 and standard deviation 1.0.</li>
  <li><code>.random2(nextDouble, X, 5)</code>: unifies X with a random number between 0 and 1, and backtracks 5 times. For example: .findall(X, .random2(nextDouble,X,5), L) will produce a list of 5 random numbers. (Similar functioning for the rest of actions: <code>nextInt</code>, <code>nextBoolean</code> or <code>nextGaussian</code> )</li>
  <li><code>.random2(nextDouble, X, 0)</code>: unifies X with a random number between 0 and 1, and but backtracks infinitely. (Similar functioning for the rest of actions: <code>nextInt</code>, <code>nextBoolean</code> or <code>nextGaussian</code> )</li>

  </ul>

  @author Francisco Grimaldo 

  @see jason.stdlib.random

*/
public class random2 extends DefaultInternalAction {
    
	private static final long serialVersionUID = 1L;

	private static InternalAction singleton = null;
    public static InternalAction create() {
        if (singleton == null) 
            singleton = new random2();
        return singleton;
    }
    
    private Random random = new Random();    
    
	public void setSeed(long s)
	{
		random.setSeed(s);
	}

    @Override public int getMinArgs() { return 2; }
    @Override public int getMaxArgs() { return 3; }
    
    @Override protected void checkArguments(Term[] args) throws JasonException {
        super.checkArguments(args); // check number of arguments
        if (!args[0].isAtom())
            throw JasonException.createWrongArgument(this,"first argument must be an atom.");                
        if (args[0].equals(ASSyntax.createAtom("setSeed")) && !args[1].isNumeric())
            throw JasonException.createWrongArgument(this," 'setSeed' needs a numeric second argument.");                
        else if (args[0].equals(ASSyntax.createAtom("nextDouble")) ||
			 args[0].equals(ASSyntax.createAtom("nextInt")) ||
		     args[0].equals(ASSyntax.createAtom("nextBoolean")) ||
			 args[0].equals(ASSyntax.createAtom("nextGaussian"))  	)
			if (args.length == 2 && !args[1].isVar())
				throw JasonException.createWrongArgument(this, args[0] + " needs a variable second argument.");
			else if (args.length == 3 && (!args[1].isVar() || !args[2].isNumeric()))
				throw JasonException.createWrongArgument(this, args[0] + " with backtracking needs a numeric second argument and a variable third argument.");
    }

    @Override
    public Object execute(final TransitionSystem ts, final Unifier un, final Term[] args) throws Exception {
        checkArguments(args);

		if (args[0].equals(ASSyntax.createAtom("setSeed"))) {
			setSeed((long) ((NumberTerm)args[1]).solve());
			return true;
		} 
		else if  (args[0].equals(ASSyntax.createAtom("nextDouble"))) 
		{ 
			if (args.length == 2)
					return un.unifies(ASSyntax.createNumber(random.nextDouble()), args[1]);
			else
			{
				final int max = (int)((NumberTerm)args[2]).solve();

				return new Iterator<Unifier>() {
					int n = 0;
					// we always have a next random number
					public boolean hasNext() { 
						return (n < max || max == 0) && ts.getUserAgArch().isRunning();  
					}
					public Unifier next() {
						Unifier c = un.clone();
						c.unifies(args[1], ASSyntax.createNumber(random.nextDouble()));
						n++;
						return c;
					}
					public void remove() {}
				};
			}
		}
		else if  (args[0].equals(ASSyntax.createAtom("nextInt"))) 
		{ 
			if (args.length == 2)
					return un.unifies(ASSyntax.createNumber(random.nextInt()), args[1]);
			else
			{
				final int max = (int)((NumberTerm)args[2]).solve();

				return new Iterator<Unifier>() {
					int n = 0;
					// we always have a next random number
					public boolean hasNext() { 
						return (n < max || max == 0) && ts.getUserAgArch().isRunning();  
					}
					public Unifier next() {
						Unifier c = un.clone();
						c.unifies(args[1], ASSyntax.createNumber(random.nextInt()));
						n++;
						return c;
					}
					public void remove() {}
				};
			}
		}
		else if  (args[0].equals(ASSyntax.createAtom("nextBoolean"))) 
		{ 
			if (args.length == 2)
					return un.unifies(ASSyntax.createAtom(Boolean.toString(random.nextBoolean())), args[1]);
			else
			{
				final int max = (int)((NumberTerm)args[2]).solve();

				return new Iterator<Unifier>() {
					int n = 0;
					// we always have a next random number
					public boolean hasNext() { 
						return (n < max || max == 0) && ts.getUserAgArch().isRunning();  
					}
					public Unifier next() {
						Unifier c = un.clone();
						c.unifies(args[1], ASSyntax.createAtom(Boolean.toString(random.nextBoolean())));
						n++;
						return c;
					}
					public void remove() {}
				};
			}
		}
		else if  (args[0].equals(ASSyntax.createAtom("nextGaussian"))) 
		{ 
			if (args.length == 2)
					return un.unifies(ASSyntax.createNumber(random.nextGaussian()), args[1]);
			else
			{
				final int max = (int)((NumberTerm)args[2]).solve();

				return new Iterator<Unifier>() {
					int n = 0;
					// we always have a next random number
					public boolean hasNext() { 
						return (n < max || max == 0) && ts.getUserAgArch().isRunning();  
					}
					public Unifier next() {
						Unifier c = un.clone();
						c.unifies(args[1], ASSyntax.createNumber(random.nextGaussian()));
						n++;
						return c;
					}
					public void remove() {}
				};
			}
		}
		else
			throw JasonException.createWrongArgument(this, args[0] + " is not a valid first argument.");
    }
	
}

